/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

 /**
 * @file l2cap_cl.c
 *
 *\author Kevin Martin <kevmarti@tik.ee.ethz.ch>
 * 
 *\date 03-22-06
 * 
 *\brief Implementation of the connection-less l2cap layer
 *
 */
#include <string.h>
#include <sys/event.h>
#include <sys/heap.h>
#include <sys/thread.h>
#include <sys/mutex.h>
#include <sys/timer.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_hci_cmds.h>
#include <bt/bt_acl_defs.h>
#include <terminal/btn-terminal.h>
#include <bt/bt_msg_defs.h>
#include <bt/l2cap_cl.h>
#include <debug/toolbox.h>

// debug settings
#define LOG_LEVEL 0
#define LOG_CLASS 4
#include <debug/log_set.h>

#if LOG_LEVEL >= 4
// string buffers for debugging
u_char _l2cap_cl_addr_buf_1[18];
u_char _l2cap_cl_addr_buf_2[18];
#endif

char _l2cap_cl_dbg_buf[3*121];

#define L2CAP_CL_TRACE_BEFORE_GEN_TSTAMP	0
#define L2CAP_CL_TRACE_AFTER_GEN_TSTAMP		1

/// Channel id of the connection-less L2CAP channel.
#define L2CAP_CL_CONLESS_CID 	0x0002

#define L2CAP_CL_TSYNC_PSM		0x3009

struct l2cap_cl_channel_s {
	struct bt_hci_acl_link* acl_link;
	u_short max_size;
	u_short written;
};

/**
 * \brief l2cap connection-less stack.
 *  
 * This struct contains all the data needed by the l2cap connection-less
 * layer itself. Basically, it holds information about the services registered
 * on the connection-less l2cap channel, as well as the information about
 * how to forward connection-oriented l2cap packets to the higher layer l2cap
 * stack.
 */
struct _l2cap_cl_stack_t {
    struct btstack* bt_stack;		///< Pointer to the lower layer bt-stack
    bt_psm_t* psmux;				///< The protocol/service multiplexor
    l2cap_cl_channel_t cl_chan;		///< Handle for the l2cap cl channel
    bt_acl_pkt_buf* pkt_buf;
    u_short acl_size;
    u_long t_arrive;
	HCI_ACL_CB;					///< Callback handling connection-oriented data
	void* acl_cb_arg;			///< Callback argument
} _l2cap_cl_stack;

/**
 * \brief Definition of the l2cap_cl pkt header.
 *
 \verbatim ---------------------------------------
 | len | cid | psm |       data         |
 ---------------------------------------
 .  2     2     2         len - 2\endverbatim
 *
 * - len: size of the data in bytes. Note: includes psm field!
 * - cid: L2CAP channel id. Set to #L2CAP_CL_CONLESS_CID for connection-less L2CAP packets.
 * - psm: Protocol/Service Multiplexer
 * 
 * See BT specification 1.2, Vol. 3, Part A, p. 34 for further details.
 */
typedef struct _l2cap_cl_pkt_hdr {
    u_char len[2];			/// size of payload (including the psm field)
    u_char cid[2];			/// channel id, set to #L2CAP_CL_CONLESS_CID
    u_char psm[L2CAP_CL_PSM_FIELD_SIZE];   	/// Protocol/Service Multiplexer
} l2cap_cl_pkt_hdr;

// this is just a dummy callback that handles packets received on different
// channels than the connectionless data channel.
struct bt_hci_pkt_acl* _l2cap_cl_cb_null(void* cb_arg,
											struct bt_hci_pkt_acl* acl_pkt,
											bt_hci_con_handle_t con_handle,
											u_char pb_flag,
											u_char bc_flag,
											u_short len,
											u_long t_arrive)
{
	DEBUG_STR("l2cap pkt rcvd! cid: %d, data:\n%s\n",
					acl_pkt->payload[2] | acl_pkt->payload[3] << 8,
					dbg_str_to_hex_str(_l2cap_cl_dbg_buf, acl_pkt->payload, len));
	return acl_pkt;
}

// this callback will be called after reception of an acl data packet
struct bt_hci_pkt_acl* _l2cap_cl_acl_cb(void* cb_arg,
										struct bt_hci_pkt_acl* pkt,
										bt_hci_con_handle_t con_handle,
										u_char pb_flag,
										u_char bc_flag,
										u_short len,
										u_long t_arrive)
{
	short retval;
	u_short cid;
	u_short psm;
	bt_acl_pkt_buf* free_buf;
    
    // get pointer to l2cap_cl packet header
    u_char* data = pkt->payload;
    l2cap_cl_pkt_hdr* hdr = (l2cap_cl_pkt_hdr*)data;

	// get channel id - if not connectionless (=0x02), call next higher layer
	cid = hdr->cid[0] | hdr->cid[1] << 8;
	if (cid != L2CAP_CL_CONLESS_CID) {
		return _l2cap_cl_stack.hci_acl_cb(cb_arg,
											pkt,
											con_handle,
											pb_flag,
											bc_flag,
											len,
											t_arrive);
	}
	
    // get con handle - if not valid, discard (return) packet
    con_handle = bt_acl_get_con_handle(pkt);
    if (con_handle == BT_HCI_HANDLE_INVALID)
    {
        ERROR("pkt with invalid hdl!\n");
        return pkt;
    }

    // else, get psm
    psm = hdr->psm[0] | hdr->psm[1] << 8;
    // get data len
    len = hdr->len[0] | hdr->len[1] << 8;
    // Note: size of PSM field is included in data size field
    // (see BT specification 1.2, Vol. 3, Part A, p. 34)
    len -= L2CAP_CL_PSM_FIELD_SIZE;
    // adjust data pointer
    data += sizeof(l2cap_cl_pkt_hdr);

	DEBUG("acl cb called! handle: %d, psm: %.4x, len: %d\n",
					con_handle, psm, len);
	DEBUG_STR("data: %s\n", dbg_str_to_hex_str(_l2cap_cl_dbg_buf, data, len));
	
	// store arrival time & deliver msg
	_l2cap_cl_stack.pkt_buf->tstamp = t_arrive;
	_l2cap_cl_stack.pkt_buf->pkt = pkt;
	retval = bt_psm_deliver_msg(_l2cap_cl_stack.psmux,
								_l2cap_cl_stack.pkt_buf,
								data,
								len,
								psm,
								&free_buf);

	if (retval)
	{
		WARNING("delivering packet to service with psm %d! code: %d!\n",
						psm, retval);
		return pkt;
	}
	// else, we store the pointer to the (empty) pkt buf obtained and return
	// the free packet it contains
	_l2cap_cl_stack.pkt_buf = free_buf;
	return free_buf->pkt;
}

long l2cap_cl_channel_open(bt_hci_con_handle_t hdl,
							u_short psm,
							u_short len,
							l2cap_cl_channel_t** cl_chan_pp)
{
	long result;
	l2cap_cl_pkt_hdr hdr;
	struct bt_hci_acl_link* acl_link;
	
	/* build the packet header */
    // Note: size of PSM field has to be added to payload size
    // (see BT specification 1.2, Vol. 3, Part A, p. 34)
    hdr.len[0] = BTN_LO(len + L2CAP_CL_PSM_FIELD_SIZE);
    hdr.len[1] = BTN_HI(len + L2CAP_CL_PSM_FIELD_SIZE);
    // write connection-less channel-id to pkt
    hdr.cid[0] = BTN_LO(L2CAP_CL_CONLESS_CID);
    hdr.cid[1] = BTN_HI(L2CAP_CL_CONLESS_CID);
    // write psm to packet
    hdr.psm[0] = BTN_LO(psm);
    hdr.psm[1] = BTN_HI(psm);
    
    /* get access to the desired acl link */
	result = bt_acl_link_open(_l2cap_cl_stack.bt_stack,
								hdl,
								HCI_PB_FIRST_FRAGMENT,
								HCI_BC_UNICAST,
                        		len + sizeof(l2cap_cl_pkt_hdr),
                        		&acl_link);
    // if error, return
	if (result < 0) return result;

    // send header
    result = bt_acl_link_write(acl_link, (u_char*)&hdr, sizeof(l2cap_cl_pkt_hdr));
    // if error, close acl link and return
    if (result < 0) {
	    bt_acl_link_close(acl_link);
	    return result;
    }
    // else, give access to the connectionless channel, and return
    *cl_chan_pp = &_l2cap_cl_stack.cl_chan;
    (*cl_chan_pp)->acl_link = acl_link;
    (*cl_chan_pp)->max_size = len;
    (*cl_chan_pp)->written = 0;
    return result;
}	

inline void l2cap_cl_channel_close(l2cap_cl_channel_t* chan)
{
	bt_acl_link_close(chan->acl_link);
}

inline long l2cap_cl_channel_write(l2cap_cl_channel_t* chan, u_char* data, u_short len)
{
	return bt_acl_link_write(chan->acl_link, data, len);
}

inline u_char* l2cap_cl_get_data_ptr(struct bt_hci_pkt_acl* pkt)
{
	return bt_acl_get_data_pointer(pkt) + sizeof(l2cap_cl_pkt_hdr);
}

inline u_short l2cap_cl_max_payload(void)
{
	return _l2cap_cl_stack.acl_size - sizeof(l2cap_cl_pkt_hdr);
}

inline u_long l2cap_cl_get_arrival_time(bt_msg* msg)
{
	return msg->tstamp;
}

/**
 * Sends a packet over the link corresponding to the handle passed. On the
 * receiving side, the incoming packet will be handled by the service that
 * is registered at the psm specified.
 */
long l2cap_cl_send(u_char* data,
					u_short len,
					bt_hci_con_handle_t hdl,
					u_short psm)
{
	long result;
	l2cap_cl_pkt_hdr hdr;
	struct bt_hci_acl_link* acl_link;

    // build l2cap_cl packet header
    // write connection-less channel-id to pkt
    hdr.cid[0] = BTN_LO(L2CAP_CL_CONLESS_CID);
    hdr.cid[1] = BTN_HI(L2CAP_CL_CONLESS_CID);
    // write payload size
    // Note: size of PSM field has to be added to payload size
    // (see BT specification 1.2, Vol. 3, Part A, p. 34)
	hdr.len[0] = BTN_LO(len + L2CAP_CL_PSM_FIELD_SIZE);
	hdr.len[1] = BTN_HI(len + L2CAP_CL_PSM_FIELD_SIZE);
    // write psm to packet
    hdr.psm[0] = BTN_LO(psm);
    hdr.psm[1] = BTN_HI(psm);

	// get access to the desired acl link
	result = bt_acl_link_open(_l2cap_cl_stack.bt_stack,
							hdl,
							HCI_PB_FIRST_FRAGMENT,
							HCI_BC_UNICAST,
							sizeof(l2cap_cl_pkt_hdr) + len,
							&acl_link);
	if (result) {
		ERROR("opening acl link (%d)!\n", result);
		return result;
	}
		
	// write l2cap cl pkt header
	result = bt_acl_link_write(acl_link, (u_char*)&hdr, sizeof(hdr));
	if (result) {
		ERROR("sending header (%d)!\n", result);
		return result;
	}
	
	// write data
	result = bt_acl_link_write(acl_link, data, len);
	if (result) {
		ERROR("sending data (%d)!\n", result);
		return result;
	}
	
	// close link
	bt_acl_link_close(acl_link);
	return result;
}

/**
 * Initializes the connection-less l2cap protocol stack
 */
void l2cap_cl_init(struct btstack* bt_stack, bt_psm_t* mux)
{
	u_short acl_size;
	struct bt_hci_pkt_acl* first_pkt;
    long retval;
    int host_buffer_size;

    // store pointer to bt-stack
    _l2cap_cl_stack.bt_stack = bt_stack;
    // store pointer to the protocol/service multiplexor
    _l2cap_cl_stack.psmux = mux;

	// initialize first packet buffer
	_l2cap_cl_stack.pkt_buf = NutHeapAllocClear(sizeof(bt_acl_pkt_buf));
	
	// register acl data callback & give first acl pkt buffer
	acl_size = bt_acl_get_max_pkt_len(bt_stack);
	first_pkt = NutHeapAllocClear(BT_ACL_HCI_PKT_HDR_LEN + acl_size);
    bt_hci_register_acl_cb(bt_stack, _l2cap_cl_acl_cb, first_pkt, NULL);

	// register dummy callback for connection oriented data
	_l2cap_cl_stack.hci_acl_cb = _l2cap_cl_cb_null;
	_l2cap_cl_stack.acl_cb_arg = NULL;
	
	// determine host_buffer_size
    host_buffer_size = BT_ACL_HCI_PKT_HDR_LEN + acl_size;
#if (BT_RADIO_TYPE == ERICSSON)
    //ericsson module demands at least 339 bytes
    if (host_buffer_size < 339)
        host_buffer_size = 339;
#endif

    //set number of buffer to module for flowcontrol!
    retval = bt_hci_host_buffer_size(bt_stack, BT_HCI_SYNC, host_buffer_size, 0, bt_psm_get_nr_default_bufs(mux), 0);
    if (retval < 0)
        DEBUG("could not set host buffer size %d buffer a %d bytes\n", bt_psm_get_nr_default_bufs(mux), host_buffer_size);
}

void l2cap_cl_register_acl_cb(struct btstack* bt_stack, HCI_ACL_CB, void* arg)
{
	_l2cap_cl_stack.hci_acl_cb = hci_acl_cb;
	_l2cap_cl_stack.acl_cb_arg = arg;
}
