/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * @file mhop_cl.h
 *
 *\author Kevin Martin <kevmarti@tik.ee.ethz.ch>
 * 
 *\date 2006/03/23
 * 
 * \brief Implementation of a connectionless multi-hop layer,
 * using the L2CAP connectionless data channel.
 * 
 * Higher layer services can use the function #mhop_cl_send_pkt to send
 * connection-less multi-hop packets to a specific specific service
 * on the target device. To send packets from hop to hop, the multi-hop
 * layer encapsulates multi-hop data into l2cap connection-less packets
 * (see #l2cap_cl.h).
 * 
 * Receiving connection-less multi-hop data packets is the same as receiving
 * point-to-point data packets: higher layer services
 * have to register a data reception callback function at the
 * protocol/service multiplexor (see #bt_psm_service_register, #l2cap_cl.h).
 * 
 * Forwarding of a packet to its destination node is done by broadcasting or by
 * consulting forwarding tables maintained by each node in the network:
 * if a destination can't be found in the forwarding table, the packet is
 * broadcasted to all directly connected devices, except to the directly
 * connected device the packet was received from.
 * 
 * On the other hand, routing (i.e. updating the forwarding tables on each node)
 * is done by storing the source node's address of an incoming packet together
 * with the connection handle the packet was received on in the forwarding
 * table (source recording). Incoming packets having a target address that
 * match to one of the entries in the forwarding table are then forwarded
 * directely to the corresponding connection handle - instead of being
 * broadcasted.
 */
#ifndef _MHOP_CL_H_
#define _MHOP_CL_H_

#include <bt/bt_defs.h>
#include <bt/bt_hci_defs.h>
#include <bt/bt_psm.h>

/// maximal size of the forwarding table
#define MHOP_CL_MAX_FWD_TABLE_SIZE				10

/// disables source recording
#define MHOP_CL_SRC_REC_OFF						0
/// enables source recording
#define MHOP_CL_SRC_REC_ON						1

/// disables pkt verbosity
#define MHOP_CL_VERBOSITY_OFF					0
/// enables pkt verbosity
#define MHOP_CL_VERBOSITY_ON					1

/// infinite time to live
#define TTL_INFINITE							0

//#define MHOP_CL_MY_ADDR (_mhop_cl_stack.my_addr)

/// Maximal payload size of a multi-hop packet
//#define MHOP_CL_PKT_MAX_PAYLOAD (L2CAP_CL_PKT_MAX_PAYLOAD - MHOP_CL_PKT_HDR_SIZE)

#define MHOP_CL_PKT_HDR_SIZE 18

/// Payload overflow error
#define MHOP_CL_ERR_PAYLOAD_OVERFLOW			-1

/**
 * \brief Sends a multi-hop packet to the psm of a remote device
 * 
 * Use this function to send a packet to a node connected via several relay
 * hops. For that the packet can be routed properly to the destination
 * specified, some additional routing information is added to the packet,
 * thus reducing the maximal payload size of the packet to
 * #MHOP_CL_MAX_MHOP_PKT_PAYLOAD.
 * 
 * Basically, usage of this function is equivalent to #l2cap_cl_send_pkt(), but
 * some additional parameters have to be given:
 * - source recording: if set to #MHOP_CL_SRC_REC_ON, intermediate nodes
 * store the connection handle the packet was received on together with the
 * bt address of the sender in their forwarding tables. Future incoming packets
 * having that address as their destination address are forwarded directely
 * over the connection handle found in the corresponding table entry.
 * - verbosity: if set to #MHOP_CL_VERBOSITY_ON, intermediate nodes are
 * signalled about the arrival of a 'verbose' packet although they are not
 * the desired destination node.
 * - ttl (time to live): after the packet has traversed 'ttl' hops, it won't
 * be forwarded any further, although the desired destination probably has not
 * been reached yet.
 * - broadcasting: broadcasting is done by sending a packet to the null-address.
 * 
 * E.g.
\verbatim
u_char data[12];
strcpy(data, "hello world");
mhop_cl_send_pkt(data, 12, #MHOP_CL_SRC_REC_ON, #MHOP_CL_VERBOSITY_ON, 20, <target_addr>, <target_psm>);
\endverbatim
 * sends "hello world" to the device with address 'target_addr', where it
 * will be delivered to the service registered at psm 'target_psm'.
 * On intermediate hops:
 * - the address of the sending device will be stored (since source recording is on),
 * - but the packet will not be delivered to the service identified by
 * 'target_psm' (since verbosity is off).
 * 
 * \param stack Pointer to the connectionless l2cap stack
 * \param pkt Pointer to the packet to send
 * \param len Payload size
 * \param src_rec To enable/disable source recording on intermediate nodes
 * \param verbosity To enable/disable packet delivery on intermediate nodes
 * \param ttl Maximal time to live for the packet - pass TTL_INFINITE
 * for inifinite time to live
 * \param dest Destination address
 * \param psm PSM the packet is delivered to on the receiving node
 * 
 * \return 0 or possible errors:
 * - #MHOP_CL_ERR_PAYLOAD_OVERFLOW
 * - #BT_ERR_BT_MODULE
 * - #BT_ERR_NO_CON
 */
long mhop_cl_send_pkt(u_char* data,
						u_short len,
						u_char src_rec,
						u_char verbose,
						u_char ttl,
						bt_addr_t dest,
						u_short psm);

/**
 * \brief broadcasts a source deletion packet
 * 
 * Sends an packet indicating that this node should no longer be considered
 * as a source node, that is, receiving nodes remove the sender of a source
 * deletion packet from the forwarding table.
 * 
 * \param stack Pointer to the connectionless l2cap stack
 * \param ttl Maximal time to live for this source deletion packet
 */
long mhop_cl_send_src_del_pkt(u_char  ttl);

/**
 * \brief Returns the source address of a received packet.
 */
u_char* mhop_cl_get_source_addr(struct bt_hci_pkt_acl* pkt);

/**
 * \brief Returns the target address of an received packet.
 */ 
u_char* mhop_cl_get_target_addr(struct bt_hci_pkt_acl* pkt);

/**
 * \brief Returns the maximal amount of data that can be sent within a
 * single multi-hop packet.
 */
u_short mhop_cl_max_payload_size(void);

/**
 * \brief Returns the bluetooth address of this device.
 * 
 * TODO: move this function to bt_hci.c
 */
u_char* mhop_cl_get_my_addr(void);

/**
 * \brief Initializes the connection-less multi-hop layer.
 * 
 * To send multi-hop data from hop to hop the packet is encapsulated
 * into a connection-less l2cap packet which is sent to the PSM
 * \p mhop_psm on the neighbor device.
 * 
 * To make use of the buffering services provided by the protocol/service
 * multiplexor (see #bt_psm.h), the multi-hop layer implements packet
 * forwarding as a service registered at PSM \p fwd_psm at the
 * protocol/service multiplexor.
 * 
 * \param stack Pointer to the bt stack to use
 * \param psmux Pointer to the protocol/service multiplexor
 * \param mhop_psm PSM the multi-hop layer should use
 * \param fwd_psm PSM the multi-hop layer should use for packet forwarding
 * \param buf_size Size of the forwarding queue
 * \param HCI_CON_TABLE_CB_REGISTRATION
 * 
 * \see #bt_psm.h
 */
void mhop_cl_init(struct btstack* bt_stack,
					bt_psm_t* psmux,
					u_short mhop_psm,
					u_short fwd_psm,
					u_short buf_size,
					HCI_CON_TABLE_CB_REGISTRATION);

#endif //_MHOP_CL_H_
